@@ -13,7 +13,7 @@ from data_product_sdk.cortex_databricks.data_platform.kafka_utils import (
 from data_product_sdk.cortex_databricks.data_platform.processing_utils import (
     process_dataframe,
 )
-from data_product_sdk.cortex_databricks.data_platform.profilling_utils import (
+from data_product_sdk.cortex_databricks.data_platform.profiling_utils import (
     calculate_profilling,
 )
 from data_product_sdk.cortex_databricks.data_platform.schema_utils import (
@@ -47,6 +47,18 @@ SILVER_DB_NAME = f"cortex_databricks_silver_{ENVIRONMENT}"
 GOLD_DB_NAME = f"cortex_databricks_gold_{ENVIRONMENT}"
 
 
+def __verify_parameters(function):
+    def wrapper(*args, **kwargs):
+        insert_modes_avalables = ["append", "upsert"]
+        if kwargs.get("insert_mode") in insert_modes_avalables:
+             return function(*args, **kwargs)
+        else:
+            raise ValueError(
+                f"Argument insert_mode is '{kwargs.get('insert_mode')}' but should be one of {insert_modes_avalables}"
+            )
+    return wrapper
+
+
 def __get_partition_type(partition_value: str):
     try:
         datetime.datetime.strptime(partition_value, "%Y-%m-%d")
@@ -117,12 +129,15 @@ def build_table_path(bucket_name: str, data_pack_name: str, dataset_name: str):
 
 
 def build_checkpoint_path(
-    bucket_name: str,
-    data_pack_name: str,
-    dataset_name: str,
-    input_table_name: str = "",
-    checkpoint_path=None,
+    bucket_name: str, data_pack_name: str, dataset_name: str, input_table_name: str = ""
 ):
+    """
+    :param bucket_name: bucket in S3 where the checkpoint is stored
+    :param data_pack_name: name of the data_pack
+    :param dataset_name: name of the dataset
+    :param input_table_name: input_table if exists
+    :return:
+    """
     assert bucket_name == BRONZE_BUCKET or (
         input_table_name and bucket_name in (SILVER_BUCKET, GOLD_BUCKET)
     ), f"bucket name must be one of {BRONZE_BUCKET}, {SILVER_BUCKET} or {GOLD_BUCKET} and {input_table_name} must be valid when using {SILVER_BUCKET} or {GOLD_BUCKET}"
@@ -131,26 +146,30 @@ def build_checkpoint_path(
     return checkpoint_location
 
 
-def read_table(spark_session: SparkSession, db_name: str, table_name: str) -> DataFrame:
+def read_table(spark_session: SparkSession, db_name: str, table_name: str, where=None) -> DataFrame:
     table_name = f"{db_name}.{table_name}"
-    df = spark_session.readStream.format("delta").table(tableName=table_name)
+
+    if where is None:
+        df = spark_session.readStream.format('delta').table(tableName=table_name)
+    else:
+        df = spark_session.readStream.format('delta').table(tableName=table_name).filter(where)
 
     return df
 
-
 def read_data(
     spark_session: SparkSession,
     data_input_path: str,
-    js_schema: dict,
+    dict_schema_or_struct_schema: dict,
     data_type: str,
     read_options: dict = None,
     infer_schema=False,
     data_pack_name=None,
     dataset_name=None,
 ):
-    assert (
-        js_schema or infer_schema
-    ), f"js_schema not provided while infer_schema is {infer_schema}. Please provide a schema or turn schema inference on"
+    if type(dict_schema_or_struct_schema) == dict:
+        assert (
+            dict_schema_or_struct_schema or infer_schema
+        ), f"js_schema not provided while infer_schema is {infer_schema}. Please provide a schema or turn schema inference on"
     if read_options is None:
         read_options = dict()
 
@@ -159,8 +178,8 @@ def read_data(
         .option("cloudFiles.format", data_type)
         .options(**read_options)
     )
-    if infer_schema:
-        schema_hints = __build_schema_hints(js_schema=js_schema)
+    if infer_schema and type(dict_schema_or_struct_schema) == dict:
+        schema_hints = __build_schema_hints(js_schema=dict_schema_or_struct_schema)
         schema_location = os.path.dirname(data_input_path)
         schema_location = os.path.join(schema_location, "_databricks_autoloader_schema")
         schema_location = os.path.join(schema_location, data_pack_name)
@@ -170,7 +189,7 @@ def read_data(
             "cloudFiles.schemaHints": schema_hints,
         }
         df = df.options(**schema_options)
-    else:
+    if type(dict_schema_or_struct_schema) == dict:
         partitions = __list_partition_columns(data_input_path)
         for partition, partition_type in partitions.items():
             partition = {
@@ -179,44 +198,24 @@ def read_data(
                 "nullable": True,
                 "type": partition_type,
             }
-            js_schema["fields"].append(partition)
-        schema = StructType.fromJson(js_schema)
-        df = df.schema(schema)
-
-    df = df.load(path=data_input_path)
-
-    return df
-
-
-def read_data_schema(
-    spark_session: SparkSession,
-    data_input_path: str,
-    data_type: str,
-    schema,
-    read_options: dict = None,
-):
-    if read_options is None:
-        read_options = dict()
-
-    df = (
-        spark_session.readStream.format("cloudFiles")
-        .option("cloudFiles.format", data_type)
-        .options(**read_options)
-    )
+            dict_schema_or_struct_schema["fields"].append(partition)
+        schema = StructType.fromJson(dict_schema_or_struct_schema)
+    else:
+        schema = dict_schema_or_struct_schema
 
     df = df.schema(schema)
-    df = df.load(path=data_input_path)
 
-    return df
+    return df.load(path=data_input_path)
 
 
-def __upsert_function(
+def __write_function(
     micro_batch_df: DataFrame,
-    batch_id: int,
     db_name,
     table_name,
+    insert_mode: str,
     sort_fields: list,
     upsert_fields: list = None,
+    partition_columns: list = None,
     spark=None,
 ):
     if spark is None:
@@ -236,10 +235,29 @@ def __upsert_function(
         [f"{current_df_name}.{i} = {micro_batch_df_name}.{i}" for i in upsert_fields]
     )
     table_name = f"{db_name}.{table_name}"
-    current_df = DeltaTable.forName(spark, table_name)
-    current_df.alias(current_df_name).merge(
-        micro_batch_df.alias(micro_batch_df_name), upsert_condition
-    ).whenMatchedUpdateAll(match_condition).whenNotMatchedInsertAll().execute()
+
+    if insert_mode == "upsert":
+        try:
+            current_df = DeltaTable.forName(spark, table_name)
+        except Exception as e:
+            current_df = None
+    else:
+       current_df = None
+
+    if current_df:
+        current_df.alias(current_df_name).merge(
+            micro_batch_df.alias(micro_batch_df_name), upsert_condition
+        ).whenMatchedUpdateAll(match_condition).whenNotMatchedInsertAll().execute()
+    else:
+        micro_batch_df.write.option("header", "true").option(
+            "mergeSchema", "true"
+        ).partitionBy(*partition_columns).format("delta").mode("append").saveAsTable(
+            table_name
+        )
+        spark.sql(f"ALTER TABLE {table_name} OWNER TO `data-platform`")
+        print(f"Created table {table_name} and loaded data")
+
+    return True
 
 
 # TODO: ADICIONAR PARAMETRO DE PATH
@@ -248,6 +266,7 @@ def write_table(
     checkpoint_location: str,
     db_name: str,
     table_name: str,
+    insert_mode: str,
     spark=None,
     partition_columns=None,
     write_options: dict = None,
@@ -277,15 +296,16 @@ def write_table(
                 f"TABELA EXISTE - upsert_fields: {upsert_fields} - sort_fields: {sort_fields}"
             )
             upsert_func = partial(
-                __upsert_function,
+                __write_function,
                 db_name=db_name,
                 table_name=table_name,
+                insert_mode=insert_mode,
                 sort_fields=sort_fields,
                 upsert_fields=upsert_fields,
+                partition_columns=partition_columns,
                 spark=spark,
             )
-            stream = stream.foreachBatch(upsert_func)
-            return stream.start()
+            return stream.foreachBatch(upsert_func).start()
         else:
             stream = stream.toTable(
                 tableName=f"{db_name}.{table_name}", partitionBy=partition_columns
@@ -306,7 +326,6 @@ def write_table(
 
 
 def __apply_profiling(df: DataFrame, profiling_s3_bucket, profiling_s3_key):
-    # check if dataframe is empty
     if df.count() == 0:
         raise ValueError("Dataframe is empty, profile aborted")
     else:
@@ -364,13 +383,37 @@ def apply_profiling(df: DataFrame, data_pack_name: str, dataset_name: str):
         df=df, profiling_s3_bucket=BRONZE_BUCKET, profiling_s3_key=profiling_s3_key
     )
 
+    return True
 
+
+def set_owner_to(spark, db_table, owner: str):
+    print(f"ALTER TABLE {db_table} OWNER TO `{owner}`")
+    spark.sql(f"ALTER TABLE {db_table} OWNER TO `{owner}`")
+
+    return True
+
+
+def set_tbl_properties(spark, db_table, tbl_properties):
+    array_properties = []
+    for k in tbl_properties:
+        text = f'"{k}"= "{tbl_properties[k]}"'
+        array_properties.append(text)
+
+    properties = ",".join(array_properties)
+    print(f"ALTER TABLE {db_table} SET TBLPROPERTIES ({properties})")
+    spark.sql(f"ALTER TABLE {db_table} SET TBLPROPERTIES ({properties})")
+
+    return True
+
+
+@__verify_parameters
 def landing_to_bronze(
     data_input_path,
     data_type: str,
     data_pack_name,
     dataset_name,
-    schema_path=None,
+    insert_mode,
+    schema_path_or_struct_schema=None,
     partition_columns=None,
     spark=None,
     read_options: dict = None,
@@ -378,6 +421,9 @@ def landing_to_bronze(
     profiling=False,
     infer_schema=False,
     reset_checkpoint=False,
+    catalog_name=None,
+    db_name=None,
+    tbl_properties=None,
 ):
     """
     Perform the ingestion of data into the bronze_layer database
@@ -386,7 +432,7 @@ def landing_to_bronze(
     :param data_type: the type of the data used as the parameter cloudFiles.format. Examples: 'csv', 'parquet', 'json'
     :param data_pack_name: name of the data_pack
     :param dataset_name: name of the dataset
-    :param schema_path: path in s3 of the json containing the schema of the data if not provided, is expected 'infer_schema' to be True
+    :param schema_path_or_struct_schema: path in s3 of the json containing the schema of the data or StructType. If not provided, is expected 'infer_schema' to be True
     :param partition_columns: the columns of the data to use as partition columns
     :param spark: the SparkSession object to use
     :param read_options: the read_options dictionary to be forwarded to spark
@@ -394,6 +440,9 @@ def landing_to_bronze(
     :param profiling: flag that indicates if the data profiling will be executed
     :param infer_schema: flag that indicates if the schema will be infered by spark
     :param reset_checkpoint: flag that indicates if the Autoloader checkpoint will be deleted
+    :param catalog_name: unity catalog name
+    :param db_name: database where the data will be written
+    :param tbl_properties: properties to add to table
     :return:
     """
     if spark is None:
@@ -406,19 +455,34 @@ def landing_to_bronze(
             data_pack_name=data_pack_name,
             dataset_name=dataset_name,
         )
+    if db_name is None:
+        db_name = BRONZE_DB_NAME
+
+    if type(schema_path_or_struct_schema) == str:
+        schema = get_schema_from_s3(schema_path_or_struct_schema)
+    else:
+        schema = schema_path_or_struct_schema
+
+    if catalog_name is None:
+        catalog_name = "hive_metastore"
+
+    if os.getenv("ENVIRONMENT") != "LOCAL":
+        spark.sql(f"USE CATALOG {catalog_name}")
 
-    js_schema = get_schema_from_s3(schema_path)
     df = read_data(
         spark,
         data_input_path=data_input_path,
-        js_schema=js_schema,
+        dict_schema_or_struct_schema=schema,
         data_type=data_type,
         read_options=read_options,
         infer_schema=infer_schema,
         data_pack_name=data_pack_name,
         dataset_name=dataset_name,
     )
-    df = process_dataframe(df, js_schema=js_schema)
+
+    if type(schema) == dict:
+        df = process_dataframe(df, js_schema=schema)
+
     df = __rename_invalid_columns(df)
 
     checkpoint_location = build_checkpoint_path(
@@ -426,70 +490,41 @@ def landing_to_bronze(
         data_pack_name=data_pack_name,
         dataset_name=dataset_name,
     )
+
+    db_name = f"{catalog_name}.{db_name}"
+
     stream = write_table(
         df,
         checkpoint_location=checkpoint_location,
         db_name=BRONZE_DB_NAME,
         table_name=dataset_name,
+        insert_mode=insert_mode,
         spark=spark,
         partition_columns=partition_columns,
         write_options=write_options,
     )
 
+    db_table = f"{db_name}.{dataset_name}"
+
+    if catalog_name != "hive_metastore":
+        set_owner_to(spark, db_table, "data-production")
+
+    if tbl_properties:
+        set_tbl_properties(spark, db_table, tbl_properties)
+
     if profiling:
         stream.awaitTermination()
         df = spark.read.table(f"{BRONZE_DB_NAME}.{dataset_name}")
         apply_profiling(df=df, data_pack_name=data_pack_name, dataset_name=dataset_name)
 
+    return True
 
-def landing_to_bronze_schema(
-    data_input_path,
-    data_type: str,
-    data_pack_name,
-    dataset_name,
-    schema=None,
-    partition_columns=None,
-    spark=None,
-    read_options: dict = None,
-    write_options: dict = None,
-    profiling=False,
-    infer_schema=False,
-):
-    if spark is None:
-        spark = get_spark_session(app_name=dataset_name)
-    if partition_columns is None:
-        partition_columns = list()
-
-    df = read_data_schema(
-        spark,
-        data_input_path=data_input_path,
-        data_type=data_type,
-        schema=schema,
-        read_options=read_options,
-    )
-
-    # df = process_dataframe(df, js_schema=schema)
-    df = __rename_invalid_columns(df)
-
-    checkpoint_location = build_checkpoint_path(
-        bucket_name=BRONZE_BUCKET,
-        data_pack_name=data_pack_name,
-        dataset_name=dataset_name,
-    )
-    stream = write_table(
-        df,
-        checkpoint_location=checkpoint_location,
-        db_name=BRONZE_DB_NAME,
-        table_name=dataset_name,
-        spark=spark,
-        partition_columns=partition_columns,
-    )
-
-
+@__verify_parameters
 def bronze_to_silver(
     input_table_name,
     data_pack_name,
     dataset_name,
+    insert_mode,
     input_db_name=BRONZE_DB_NAME,
     partition_columns=None,
     spark=None,
@@ -531,6 +566,7 @@ def bronze_to_silver(
         checkpoint_location=checkpoint_location,
         db_name=SILVER_DB_NAME,
         table_name=dataset_name,
+        insert_mode=insert_mode,
         spark=spark,
         partition_columns=partition_columns,
         write_options=write_options,
@@ -542,10 +578,12 @@ def bronze_to_silver(
         apply_profiling(df=df, data_pack_name=data_pack_name, dataset_name=dataset_name)
 
 
+@__verify_parameters
 def silver_to_gold(
     input_table_name,
     data_pack_name,
     dataset_name,
+    insert_mode,
     input_db_name=SILVER_DB_NAME,
     partition_columns=None,
     spark=None,
@@ -595,6 +633,7 @@ def silver_to_gold(
         checkpoint_location=checkpoint_location,
         db_name=GOLD_DB_NAME,
         table_name=dataset_name,
+        insert_mode=insert_mode,
         spark=spark,
         partition_columns=partition_columns,
         write_options=write_options,
